@@ -1,17 +1,43 @@
+from __future__ import annotations
+
 import logging
 import sys
 import threading
 import uuid
+from typing import TYPE_CHECKING, Any, Optional, Protocol, TypeVar
 
 from pykka import ActorDeadError, ActorRef, ActorRegistry, messages
+from pykka._introspection import get_attr_directly
+
+if TYPE_CHECKING:
+    from types import TracebackType
+
+    from pykka import Future
+    from pykka._envelope import Envelope
 
 __all__ = ["Actor"]
 
+
 logger = logging.getLogger("pykka")
 
 
+A = TypeVar("A", bound="Actor")
+
+
+class ActorInbox(Protocol):
+    def put(self, envelope: Envelope[Any], /) -> None:
+        ...
+
+    def get(self) -> Envelope[Any]:
+        ...
+
+    def empty(self) -> bool:
+        ...
+
+
 class Actor:
-    """
+    """An actor is an execution unit that executes concurrently with other actors.
+
     To create an actor:
 
     1. subclass one of the :class:`Actor` implementations:
@@ -53,9 +79,14 @@ class Actor:
     """
 
     @classmethod
-    def start(cls, *args, **kwargs):
-        """
-        Start an actor and register it in the
+    def start(
+        cls: type[A],
+        *args: Any,
+        **kwargs: Any,
+    ) -> ActorRef[A]:
+        """Start an actor.
+
+        Starting an actor also registers it in the
         :class:`ActorRegistry <pykka.ActorRegistry>`.
 
         Any arguments passed to :meth:`start` will be passed on to the class
@@ -90,41 +121,60 @@ class Actor:
         )
         ActorRegistry.register(obj.actor_ref)
         logger.debug(f"Starting {obj}")
-        obj._start_actor_loop()
+        obj._start_actor_loop()  # noqa: SLF001
         return obj.actor_ref
 
     @staticmethod
-    def _create_actor_inbox():
-        """Internal method for implementors of new actor types."""
+    def _create_actor_inbox() -> ActorInbox:
+        """Create an inbox for the actor.
+
+        Internal method for implementors of new actor types.
+        """
         raise NotImplementedError("Use a subclass of Actor")
 
     @staticmethod
-    def _create_future():
-        """Internal method for implementors of new actor types."""
+    def _create_future() -> Future[Any]:
+        """Create a future for the actor.
+
+        Internal method for implementors of new actor types.
+        """
         raise NotImplementedError("Use a subclass of Actor")
 
-    def _start_actor_loop(self):
-        """Internal method for implementors of new actor types."""
+    def _start_actor_loop(self) -> None:
+        """Create and start the actor's event loop.
+
+        Internal method for implementors of new actor types.
+        """
         raise NotImplementedError("Use a subclass of Actor")
 
     #: The actor URN string is a universally unique identifier for the actor.
     #: It may be used for looking up a specific actor using
     #: :meth:`ActorRegistry.get_by_urn`.
-    actor_urn = None
+    actor_urn: str
 
     #: The actor's inbox. Use :meth:`ActorRef.tell`, :meth:`ActorRef.ask`, and
     #: friends to put messages in the inbox.
-    actor_inbox = None
+    actor_inbox: ActorInbox
 
-    #: The actor's :class:`ActorRef` instance.
-    actor_ref = None
+    _actor_ref: ActorRef[Any]
+
+    @property
+    def actor_ref(self: A) -> ActorRef[A]:
+        """The actor's :class:`ActorRef` instance."""
+        # This property only exists to improve the typing of the ActorRef.
+        return self._actor_ref
 
     #: A :class:`threading.Event` representing whether or not the actor should
     #: continue processing messages. Use :meth:`stop` to change it.
-    actor_stopped = None
+    actor_stopped: threading.Event
+
+    def __init__(
+        self,
+        *_args: Any,
+        **_kwargs: Any,
+    ) -> None:
+        """Create actor.
 
-    def __init__(self, *args, **kwargs):
-        """
         Your are free to override :meth:`__init__`, but you must call your
         superclass' :meth:`__init__` to ensure that fields :attr:`actor_urn`,
         :attr:`actor_inbox`, and :attr:`actor_ref` are initialized.
@@ -144,23 +194,20 @@ class Actor:
         self.actor_inbox = self._create_actor_inbox()
         self.actor_stopped = threading.Event()
 
-        self.actor_ref = ActorRef(self)
+        self._actor_ref = ActorRef(self)
 
-    def __str__(self):
+    def __str__(self) -> str:
         return f"{self.__class__.__name__} ({self.actor_urn})"
 
-    def stop(self):
-        """
-        Stop the actor.
+    def stop(self) -> None:
+        """Stop the actor.
 
         It's equivalent to calling :meth:`ActorRef.stop` with ``block=False``.
         """
-        self.actor_ref.tell(messages._ActorStop())
+        self.actor_ref.tell(messages._ActorStop())  # noqa: SLF001
 
-    def _stop(self):
-        """
-        Stops the actor immediately without processing the rest of the inbox.
-        """
+    def _stop(self) -> None:
+        """Stop the actor immediately without processing the rest of the inbox."""
         ActorRegistry.unregister(self.actor_ref)
         self.actor_stopped.set()
         logger.debug(f"Stopped {self}")
@@ -169,17 +216,22 @@ class Actor:
         except Exception:
             self._handle_failure(*sys.exc_info())
 
-    def _actor_loop(self):
-        """
-        The actor's event loop.
+    def _actor_loop(self) -> None:
+        """Run the actor's core loop.
 
         This is the method that will be executed by the thread or greenlet.
         """
+        self._actor_loop_setup()
+        self._actor_loop_running()
+        self._actor_loop_teardown()
+
+    def _actor_loop_setup(self) -> None:
         try:
             self.on_start()
         except Exception:
             self._handle_failure(*sys.exc_info())
 
+    def _actor_loop_running(self) -> None:
         while not self.actor_stopped.is_set():
             envelope = self.actor_inbox.get()
             try:
@@ -205,10 +257,11 @@ class Actor:
                 self._stop()
                 ActorRegistry.stop_all()
 
+    def _actor_loop_teardown(self) -> None:
         while not self.actor_inbox.empty():
             envelope = self.actor_inbox.get()
             if envelope.reply_to is not None:
-                if isinstance(envelope.message, messages._ActorStop):
+                if isinstance(envelope.message, messages._ActorStop):  # noqa: SLF001
                     envelope.reply_to.set(None)
                 else:
                     envelope.reply_to.set_exception(
@@ -222,8 +275,9 @@ class Actor:
                         )
                     )
 
-    def on_start(self):
-        """
+    def on_start(self) -> None:
+        """Run code at the beginning of the actor's life.
+
         Hook for doing any setup that should be done *after* the actor is
         started, but *before* it starts processing messages.
 
@@ -234,10 +288,10 @@ class Actor:
         If an exception is raised by this method the stack trace will be
         logged, and the actor will stop.
         """
-        pass
 
-    def on_stop(self):
-        """
+    def on_stop(self) -> None:
+        """Run code at the end of the actor's life.
+
         Hook for doing any cleanup that should be done *after* the actor has
         processed the last message, and *before* the actor stops.
 
@@ -250,19 +304,29 @@ class Actor:
         If an exception is raised by this method the stack trace will be
         logged, and the actor will stop.
         """
-        pass
 
-    def _handle_failure(self, exception_type, exception_value, traceback):
-        """Logs unexpected failures, unregisters and stops the actor."""
+    def _handle_failure(
+        self,
+        exception_type: Optional[type[BaseException]],
+        exception_value: Optional[BaseException],
+        traceback: Optional[TracebackType],
+    ) -> None:
+        """Log unexpected failures, unregisters and stops the actor."""
         logger.error(
             f"Unhandled exception in {self}:",
-            exc_info=(exception_type, exception_value, traceback),
+            exc_info=(exception_type, exception_value, traceback),  # type: ignore[arg-type]
         )
         ActorRegistry.unregister(self.actor_ref)
         self.actor_stopped.set()
 
-    def on_failure(self, exception_type, exception_value, traceback):
-        """
+    def on_failure(
+        self,
+        exception_type: Optional[type[BaseException]],
+        exception_value: Optional[BaseException],
+        traceback: Optional[TracebackType],
+    ) -> None:
+        """Run code when an unhandled exception is raised.
+
         Hook for doing any cleanup *after* an unhandled exception is raised,
         and *before* the actor stops.
 
@@ -275,27 +339,25 @@ class Actor:
         If an exception is raised by this method the stack trace will be
         logged, and the actor will stop.
         """
-        pass
 
-    def _handle_receive(self, message):
-        """Handles messages sent to the actor."""
-        if isinstance(message, messages._ActorStop):
+    def _handle_receive(self, message: Any) -> Any:
+        """Handle messages sent to the actor."""
+        if isinstance(message, messages._ActorStop):  # noqa: SLF001
             return self._stop()
         if isinstance(message, messages.ProxyCall):
-            callee = self._get_attribute_from_path(message.attr_path)
+            callee = get_attr_directly(self, message.attr_path)
             return callee(*message.args, **message.kwargs)
         if isinstance(message, messages.ProxyGetAttr):
-            attr = self._get_attribute_from_path(message.attr_path)
+            attr = get_attr_directly(self, message.attr_path)
             return attr
         if isinstance(message, messages.ProxySetAttr):
-            parent_attr = self._get_attribute_from_path(message.attr_path[:-1])
+            parent_attr = get_attr_directly(self, message.attr_path[:-1])
             attr_name = message.attr_path[-1]
             return setattr(parent_attr, attr_name, message.value)
         return self.on_receive(message)
 
-    def on_receive(self, message):
-        """
-        May be implemented for the actor to handle regular non-proxy messages.
+    def on_receive(self, message: Any) -> Any:
+        """May be implemented for the actor to handle regular non-proxy messages.
 
         :param message: the message to handle
         :type message: any
@@ -303,38 +365,3 @@ class Actor:
         :returns: anything that should be sent as a reply to the sender
         """
         logger.warning(f"Unexpected message received by {self}: {message}")
-
-    def _get_attribute_from_path(self, attr_path):
-        """
-        Traverses the path and returns the attribute at the end of the path.
-        """
-        attr = self
-        for attr_name in attr_path:
-            attr = getattr(attr, attr_name)
-        return attr
-
-    def _introspect_attribute_from_path(self, attr_path):
-        """Get attribute information from ``__dict__`` on the container."""
-        if not attr_path:
-            return self
-
-        parent = self._get_attribute_from_path(attr_path[:-1])
-        parent_attrs = self._introspect_attributes(parent)
-        attr_name = attr_path[-1]
-
-        try:
-            return parent_attrs[attr_name]
-        except KeyError:
-            raise AttributeError(
-                f"type object {parent.__class__.__name__!r} "
-                f"has no attribute {attr_name!r}"
-            )
-
-    def _introspect_attributes(self, obj):
-        """Combine ``__dict__`` from ``obj`` and all its superclasses."""
-        result = {}
-        for cls in reversed(obj.__class__.mro()):
-            result.update(cls.__dict__)
-        if hasattr(obj, "__dict__"):
-            result.update(obj.__dict__)
-        return result
